@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 
 # Urwid SàT extensions
-# Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014 Jérôme Poisson (goffi@goffi.org)
+# Copyright (C) 2009-2016 Jérôme Poisson (goffi@goffi.org)
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -22,6 +22,10 @@ import logging as log
 import encodings
 utf8decode = lambda s: encodings.codecs.utf_8_decode(s)[0]
 
+import uuid
+
+import collections
+
 from urwid.util import is_mouse_press #XXX: is_mouse_press is not included in urwid in 1.0.0
 from .keys import action_key_map as a_key
 
@@ -54,16 +58,21 @@ class AdvancedEdit(urwid.Edit):
     new behaviour: emit a 'click' signal when enter is pressed"""
     signals = urwid.Edit.signals + ['click']
 
+    def getValue(self):
+        return self.get_edit_text()
+
     def setCompletionMethod(self, callback):
         """Define method called when completion is asked
+
         @callback: method with 2 arguments:
-                    - the text to complete
+                    - the text to complete (part after cursor position is ignored)
                     - if there was already a completion, a dict with
                         - 'completed':last completion
                         - 'completion_pos': cursor position where the completion starts
                         - 'position': last completion cursor position
                       this dict must be used (and can be filled) to find next completion)
-                   and which return the full text completed"""
+                   and which return the full text completed
+        """
         self.completion_cb = callback
         self.completion_data = {}
 
@@ -215,8 +224,8 @@ class AlwaysSelectableText(urwid.WidgetWrap):
         @param selected: is the text selected ?
         """
         self.focus_attr = focus_attr
-        self.__selected = False
-        self.__was_focused = False
+        self._selected = False
+        self._was_focused = False
         self.header = header
         self.text = text
         urwid.WidgetWrap.__init__(self, urwid.Text("",align=align))
@@ -242,7 +251,7 @@ class AlwaysSelectableText(urwid.WidgetWrap):
     def set_text(self, text):
         """/!\ set_text doesn't change self.selected_txt !"""
         self.text = text
-        self.setState(self.__selected,invisible=True)
+        self.setState(self._selected,invisible=True)
 
     def setSelectedText(self, text=None):
         """Text to display when selected
@@ -250,12 +259,12 @@ class AlwaysSelectableText(urwid.WidgetWrap):
         if text == None:
             text = ('selected',self.getValue())
         self.selected_txt = text
-        if self.__selected:
-            self.setState(self.__selected)
+        if self._selected:
+            self.setState(self._selected)
 
-    def __set_txt(self):
+    def _set_txt(self):
         txt_list = [self.header]
-        txt = self.selected_txt if self.__selected else self.text
+        txt = self.selected_txt if self._selected else self.text
         if isinstance(txt,list):
             txt_list.extend(txt)
         else:
@@ -265,31 +274,33 @@ class AlwaysSelectableText(urwid.WidgetWrap):
 
     def setState(self, selected, invisible=False):
         """Change state
+
         @param selected: boolean state value
-        @param invisible: don't emit change signal if True"""
-        assert(type(selected)==bool)
-        self.__selected=selected
-        self.__set_txt()
-        self.__was_focused = False
+        @param invisible: don't emit change signal if True
+        """
+        assert type(selected)==bool
+        self._selected=selected
+        self._set_txt()
+        self._was_focused = False
         self._invalidate()
         if not invisible:
-            self._emit("change", self.__selected)
+            self._emit("change", self._selected)
 
     def getState(self):
-        return self.__selected
+        return self._selected
 
     def selectable(self):
         return True
 
     def keypress(self, size, key):
         if key in (a_key['TEXT_SELECT'], a_key['TEXT_SELECT2']):
-            self.setState(not self.__selected)
+            self.setState(not self._selected)
         else:
             return key
 
     def mouse_event(self, size, event, button, x, y, focus):
         if is_mouse_press(event) and button == 1:
-            self.setState(not self.__selected)
+            self.setState(not self._selected)
             return True
 
         return False
@@ -297,11 +308,11 @@ class AlwaysSelectableText(urwid.WidgetWrap):
     def render(self, size, focus=False):
         attr_list = self._w.base_widget._attrib
         if not focus:
-            if self.__was_focused:
-                self.__set_txt()
-                self.__was_focused = False
+            if self._was_focused:
+                self._set_txt()
+                self._was_focused = False
         else:
-            if not self.__was_focused:
+            if not self._was_focused:
                 if not attr_list:
                     attr_list.append((self.focus_attr,len(self._w.base_widget.text)))
                 else:
@@ -315,7 +326,7 @@ class AlwaysSelectableText(urwid.WidgetWrap):
                                 attr+="_focus"
                                 attr_list[idx] = (attr,attr_len)
                 self._w.base_widget._invalidate()
-                self.__was_focused = True #bloody ugly hack :)
+                self._was_focused = True #bloody ugly hack :)
         return self._w.render(size, focus)
 
 
@@ -337,11 +348,11 @@ class ClickableText(SelectableText):
 
 class CustomButton(ClickableText):
 
-    def __init__(self, label, on_press=None, user_data=None, left_border = "[ ", right_border = " ]"):
+    def __init__(self, label, on_press=None, user_data=None, left_border="[ ", right_border=" ]", align="left"):
         self.label = label
         self.left_border = left_border
         self.right_border = right_border
-        super(CustomButton, self).__init__([left_border, label, right_border])
+        super(CustomButton, self).__init__([left_border, label, right_border], align=align)
         self.size = len(self.get_text())
         if on_press:
             urwid.connect_signal(self, 'click', on_press, user_data)
@@ -359,13 +370,14 @@ class CustomButton(ClickableText):
 
 
 class ListOption(unicode):
-    """ Class similar to unicode, but which make the difference between value and label
+    """Unicode which manage label and value
+
+    This class similar to unicode, but which make the difference between value and label
     label is show when use as unicode, the .value attribute contain the actual value
     Can be initialised with:
         - basestring (label = value = given string)
         - a tuple with (value, label)
     XXX: comparaison is made against value, not the label which is the one displayed
-
     """
 
     def __new__(cls, option):
@@ -446,19 +458,119 @@ class UnselectableListBox(urwid.ListBox):
         return super(UnselectableListBox, self).render(size, focus)
 
 
-class GenericList(urwid.WidgetWrap):
+class SimpleListWalkerWithCb(urwid.SimpleListWalker):
+    """a SimpleListWalker which call callbacks on items changes"""
+
+    def __init__(self, contents, on_new=None, on_delete=None):
+        """
+        @param contents: list to copy into this object
+        @param on_new: callback to call when an item is added
+        @param on_delete: callback to call when an item is deleted
+        """
+        # XXX: we can't use modified signal as it doesn't return the modified item
+        super(SimpleListWalkerWithCb, self).__init__(contents)
+        for content in contents:
+            on_new(content)
+        self._on_new = on_new
+        self._on_delete = on_delete
+
+    def __cbSingle(self, item, cb):
+        try:
+            cb(item)
+        except TypeError:
+            pass
+
+    def __cbMulti(self, items, cb):
+        if cb is not None:
+            for item in items:
+                    cb(item)
+
+    def __add__(self, new_list):
+        self.__cbMulti(new_list, self._on_new)
+        return super(SimpleListWalkerWithCb, self).__add__(new_list)
+
+    def __delitem__(self, item):
+        self.__cbSingle(item, self._on_delete)
+        return super(SimpleListWalkerWithCb, self).__delitem__(item)
+
+    def __delslice__(self, i,j):
+        items = super(SimpleListWalkerWithCb, self).__getslice__(i,j)
+        self.__cbMulti(items, self._on_delete)
+        return super(SimpleListWalkerWithCb, self).__delslice(i,j)
+
+    def __iadd__(self, y):
+        raise NotImplementedError
+
+    def __imul__(self, y):
+        raise NotImplementedError
+
+    def  __mul__(self, n):
+        raise NotImplementedError
+
+    def  __rmul__(self, n):
+        raise NotImplementedError
+
+    def __setitem__(self, i, y):
+        parent = super(SimpleListWalkerWithCb, self)
+        self.__cbSingle(y, self._on_new)
+        to_delete = parent.__getitem__(i)
+        self.__cbSingle(to_delete, self._on_delete)
+        return parent.__setitem__(self, i, y)
+
+    def __setslice__(self, i, j, y):
+        parent = super(SimpleListWalkerWithCb, self)
+        items_to_delete = parent.__getslice__(i,j)
+        self.__cbMulti(items_to_delete, self._on_delete)
+        if hasattr(y, '__iter__'):
+            self.__cbMulti(y, self._on_new)
+        else:
+            self.__cbSingle(y, self._on_new)
+        return parent.__setslice__(i, j, y)
+
+    def append(self, obj):
+        self.__cbSingle(obj, self._on_new)
+        return super(SimpleListWalkerWithCb, self).append(obj)
+
+    def extend(self, it):
+        self.__cbMulti(it, self.__on_new)
+        return super(SimpleListWalkerWithCb, self).extend(it)
+
+    def insert(self, idx, obj):
+        self.__cbSingle(obj, self.__on_new)
+        return super(SimpleListWalkerWithCb, self).insert(idx, obj)
+
+    def pop(self, idx=None):
+        if idx is None:
+            idx=len(self)-1
+
+        parent = super(SimpleListWalkerWithCb, self)
+        to_remove = parent.__getitem__(idx)
+        self.__cbSingle(to_remove, self._on_delete)
+        return parent.pop(idx)
+
+    def remove(self, val):
+        ret = super(SimpleListWalkerWithCb, self).remove(val)
+        self.__cbSingle(val, self._on_delete)
+        return ret
+
+
+class GenericList(urwid.ListBox):
     signals = ['click','change']
 
     def __init__(self, options, style=None, align='left', option_type = SelectableText, on_click=None, on_change=None, user_data=None):
-        """
-        Widget managing list of string and their selection
+        """Widget managing list of string and their selection
+
         @param options: list of strings used for options
         @param style: list of string:
             - 'single' if only one must be selected
             - 'no_first_select' nothing selected when list is first displayed
             - 'can_select_none' if we can select nothing
         @param align: alignement of text inside the list
+        @param option_type: callable (usually a class) which will be called with:
+            - option as first argument
+            - align=align as keyword argument
         @param on_click: method called when click signal is emited
+        @param on_change: method called when change signal is emited
         @param user_data: data sent to the callback for click signal
         """
         if style is None:
@@ -476,12 +588,22 @@ class GenericList(urwid.WidgetWrap):
         if on_change:
             urwid.connect_signal(self, 'change', on_change, user_data)
 
-        self.content = urwid.SimpleListWalker([])
-        self.list_box = urwid.ListBox(self.content)
-        urwid.WidgetWrap.__init__(self, self.list_box)
+        self.content = SimpleListWalkerWithCb([], self._addSignals, lambda widget: self._emit('change'))
+        super(GenericList, self).__init__(self.content)
         self.changeValues(options)
 
-    def __onStateChange(self, widget, selected):
+    def _addSignals(self, widget):
+        for signal, callback in (('change', self._onStateChange), ('click', self._onClick)):
+            try:
+                urwid.connect_signal(widget, signal, callback)
+            except NameError:
+                pass #the widget given doesn't support the signal
+
+    @property
+    def contents(self):
+        return self.content
+
+    def _onStateChange(self, widget, selected, *args):
         if self.single:
             if not selected and not self.can_select_none:
                 #if in single mode, it's forbidden to unselect a value
@@ -490,10 +612,13 @@ class GenericList(urwid.WidgetWrap):
             if selected:
                 self.unselectAll(invisible=True)
                 widget.setState(True, invisible=True)
-        self._emit("change")
+        self._emit("change", widget, selected, *args)
 
-    def __onClick(self, widget):
-        self._emit("click", widget)
+    def _onClick(self, widget, *args):
+        if widget not in self.content:
+            urwid.disconnect_signal(widget, "click", self._onClick)
+            return
+        self._emit("click", widget, *args)
 
     def unselectAll(self, invisible=False):
         for widget in self.content:
@@ -527,9 +652,6 @@ class GenericList(urwid.WidgetWrap):
                 result.append(widget.getValue())
         return result
 
-    def getDisplayWidget(self):
-        return self.list_box
-
     def changeValues(self, new_values):
         """Change all values in one shot"""
         new_values = ListOption.fromOptions(new_values)
@@ -537,20 +659,13 @@ class GenericList(urwid.WidgetWrap):
             old_selected = self.getSelectedValues()
         widgets = []
         for option in new_values:
-            widget = self.option_type(option, self.align)
+            widget = self.option_type(option, align=self.align)
             if not self.first_display and option in old_selected:
                 widget.setState(True)
             widgets.append(widget)
-            for signal, callback in (('change', self.__onStateChange), ('click', self.__onClick)):
-                try:
-                    urwid.connect_signal(widget, signal, callback)
-                except NameError:
-                    pass #the widget given doesn't support the signal
         self.content[:] = widgets
         if self.first_display and self.single and new_values and not self.no_first_select:
             self.content[0].setState(True)
-        display_widget = self.getDisplayWidget()
-        self._set_w(display_widget)
         self._emit('change')
         self.first_display = False
 
@@ -558,8 +673,10 @@ class GenericList(urwid.WidgetWrap):
         """Select the first item which has the given value.
 
         @param value
-        @param move_focus (boolean): True to move the focus on the selected value,
-                                     False to leave the focus position unchanged.
+        @param move_focus (bool):
+            - True to move the focus on the selected value,
+            - False to leave the focus position unchanged.
+
         """
         self.unselectAll()
         idx = 0
@@ -567,7 +684,7 @@ class GenericList(urwid.WidgetWrap):
             if widget.getValue() == value:
                 widget.setState(True)
                 if move_focus:
-                    self.list_box.focus_position = idx
+                    self.focus_position = idx
                 return
             idx+=1
 
@@ -589,7 +706,7 @@ class GenericList(urwid.WidgetWrap):
                 if widget.getValue() == value:
                     widget.setState(True)
                     if move_focus:
-                        self.list_box.focus_position = idx
+                        self.focus_position = idx
                 idx += 1
 
 
@@ -602,19 +719,19 @@ class List(urwid.Widget):
         if style is None:
             style = []
         self.genericList = GenericList(options, style, align, option_type, on_click, on_change, user_data)
-        urwid.connect_signal(self.genericList, 'change', self._onChange)
-        urwid.connect_signal(self.genericList, 'click', self._onClick)
+        urwid.connect_signal(self.genericList, 'change', lambda *args: self._emit('change'))
         self.max_height = max_height
 
-    def _onChange(self, widget):
-        self._emit('change')
-
-    def _onClick(self, widget):
-        self._emit('click')
+    @property
+    def contents(self):
+        return self.genericList.content
 
     def selectable(self):
         return True
 
+    def get_cursor_coords(self, size):
+        return self.genericList.get_cursor_coords((size[0], self._getHeight(size, True)))
+
     def keypress(self, size, key):
         return self.displayWidget(size,True).keypress(size, key)
 
@@ -648,10 +765,13 @@ class List(urwid.Widget):
     def rows(self, size, focus=False):
         return self.displayWidget(size, focus).rows(size, focus)
 
-    def displayWidget(self, size, focus):
+    def _getHeight(self, size, focus):
         list_size = sum([wid.rows(size, focus) for wid in self.genericList.content])
         height = min(list_size,self.max_height) or 1
-        return urwid.BoxAdapter(self.genericList, height)
+        return height
+
+    def displayWidget(self, size, focus):
+        return urwid.BoxAdapter(self.genericList, self._getHeight(size, focus))
 
 
 ## MISC ##
@@ -669,7 +789,7 @@ class NotificationBar(urwid.WidgetWrap):
         urwid.WidgetWrap.__init__(self, urwid.AttrMap(self.columns,'notifs'))
         self.notifs = []
 
-    def __modQueue(self):
+    def _modQueue(self):
         """must be called each time the notifications queue is changed"""
         self.waitNotifs.set_text(('notifs',"(%i)" % len(self.notifs) if self.notifs else ''))
         self._emit('change')
@@ -687,7 +807,17 @@ class NotificationBar(urwid.WidgetWrap):
     def addPopUp(self, pop_up_widget):
         """Add a popup to the waiting queue"""
         self.notifs.append(('popup',pop_up_widget))
-        self.__modQueue()
+        self._modQueue()
+
+    def removePopUp(self, pop_up_widget):
+        """Remove a popup from the waiting queue"""
+        for idx, (wid_type, widget) in enumerate(self.notifs):
+            if widget == pop_up_widget:
+                del self.notifs[idx]
+                self._modQueue()
+                return
+
+        raise ValueError(u"trying to remove an unknown pop_up_widget")
 
     def addMessage(self, message):
         "Add a message to the notificatio bar"
@@ -697,7 +827,7 @@ class NotificationBar(urwid.WidgetWrap):
             self._emit('change')
         else:
             self.notifs.append(('message',message))
-            self.__modQueue()
+            self._modQueue()
 
     def showNext(self):
         """Show next message if any, else delete current message"""
@@ -709,7 +839,7 @@ class NotificationBar(urwid.WidgetWrap):
         if found:
             self.notifs.remove(found)
             self.message.set_text(('notifs',found[1]))
-            self.__modQueue()
+            self._modQueue()
             self.focus_possition = 1
         else:
             self.message.set_text('')
@@ -725,7 +855,7 @@ class NotificationBar(urwid.WidgetWrap):
                 break
         if ret:
             self.notifs.remove(notif)
-            self.__modQueue()
+            self._modQueue()
         return ret
 
     def isQueueEmpty(self):
@@ -868,75 +998,125 @@ class Menu(urwid.WidgetWrap):
         self.__buildOverlay(button.get_label(),
                             self.x_orig + self._w.base_widget.getStartCol(button))
 
+MenuItem = collections.namedtuple('MenuItem', ('name', 'widget'))
 
 class MenuRoller(urwid.WidgetWrap):
 
-    def __init__(self,menus_list):
+    def __init__(self, menus_list):
         """Create a MenuRoller
-        @param menus_list: list of tuple with (name, Menu_instance), name can be None
+
+        @param menus_list: list of tuples which can be either:
+            - (name, Menu instance)
+            - (name, Menu instance, id)
         """
-        assert (menus_list)
-        self.selected = 0
-        self.name_list = []
-        self.menus = {}
+        assert menus_list
+        self.selected = None
+        self.menu_items = collections.OrderedDict()
 
         self.columns = urwid.Columns([urwid.Text(''),urwid.Text('')])
         urwid.WidgetWrap.__init__(self, self.columns)
 
         for menu_tuple in menus_list:
-            name,menu = menu_tuple
-            self.addMenu(name, menu)
+            try:
+                name, menu, id_ = menu_tuple
+            except ValueError:
+                name, menu = menu_tuple
+                id_ = None
+            self.addMenu(name, menu, id_)
 
     def _showSelected(self):
         """show menu selected"""
-        name_txt = u'\u21c9 '+self.name_list[self.selected]+u' \u21c7 '
-        current_name = ClickableText(name_txt)
-        name_len = len(name_txt)
-        current_menu = self.menus[self.name_list[self.selected]]
-        current_menu.setOrigX(name_len)
-        self.columns.contents[0] = (current_name, ('given', name_len, False))
-        self.columns.contents[1] = (current_menu, ('weight', 1, False))
+        if self.selected is None:
+            self.columns.contents[0] = (urwid.Text(''), ('given', 0, False))
+            self.columns.contents[1] = (urwid.Text(''), ('weight', 1, False))
+        else:
+            menu_item = self.menu_items[self.selected]
+            name_txt = u'\u21c9 ' + menu_item.name + u' \u21c7 '
+            current_name = ClickableText(name_txt)
+            name_len = len(name_txt)
+            current_menu = menu_item.widget
+            current_menu.setOrigX(name_len)
+            self.columns.contents[0] = (current_name, ('given', name_len, False))
+            self.columns.contents[1] = (current_menu, ('weight', 1, False))
 
     def keypress(self, size, key):
+        menu_ids = self.menu_items.keys()
+        try:
+            idx = menu_ids.index(self.selected)
+        except ValueError:
+            return super(MenuRoller, self).keypress(size, key)
+
         if key==a_key['MENU_ROLLER_UP']:
             if self.columns.get_focus_column()==0:
-                if self.selected > 0:
-                    self.selected -= 1
+                if idx > 0:
+                    self.selected = menu_ids[idx-1]
                     self._showSelected()
                 return
         elif key==a_key['MENU_ROLLER_DOWN']:
             if self.columns.get_focus_column()==0:
-                if self.selected < len(self.name_list)-1:
-                    self.selected += 1
+                if idx < len(menu_ids)-1:
+                    self.selected = menu_ids[idx+1]
                     self._showSelected()
                 return
         elif key==a_key['MENU_ROLLER_RIGHT']:
             if self.columns.get_focus_column()==0 and \
                 (isinstance(self.columns.contents[1][0], urwid.Text) or \
-                self.menus[self.name_list[self.selected]].getMenuSize()==0):
+                self.menu_items[self.selected].widget.getMenuSize()==0):
                 return #if we have no menu or the menu is empty, we don't go the right column
 
         return super(MenuRoller, self).keypress(size, key)
 
-    def addMenu(self, name_param, menu):
-        name = name_param or ''
-        if name not in self.name_list:
-            self.name_list.append(name)
-        self.menus[name] = menu
-        if self.name_list[self.selected] == name:
+    def addMenu(self, name, widget, menu_id=None):
+        """Add a menu
+
+        @param name: name of the menu to add, it name already exists, menu is not added
+        @param widget: instance of Menu
+        @param menu_id: id to use of this menu, or None to generate
+        @return: menu_id
+        """
+        names = {menu_item.name: id_ for id_, menu_item in self.menu_items.iteritems()}
+
+        if name not in names:
+            id_ = menu_id or str(uuid.uuid4())
+            if id_ in self.menu_items:
+                raise ValueError('Conflict: the id [{}] is already used'.format(id_))
+            self.menu_items[id_] = MenuItem(name, widget)
+        else:
+            id_ = names[name]
+            menu_item = self.menu_items[id_]
+            if menu_item.widget is not widget:
+                raise ValueError("The menu with id [{}] exists and doesn't contain the given instance. Use replaceMenu if you want to change the menu.".format(id_))
+        if self.selected is None:
+            self.selected = id_
+            self._showSelected()
+        return id_
+
+    def replaceMenu(self, name, widget, menu_id):
+        """Add a menu or replace it if the id already exists
+
+        @param name: name of the menu to add, it name already exists, menu is not added
+        @param widget: instance of Menu
+        @param menu_id: id or the menu
+        """
+        assert menu_id is not None
+        if menu_id in self.menu_items:
+            del self.menu_items[menu_id]
+        self.addMenu(name, widget, menu_id)
+        if self.selected == menu_id:
             self._showSelected() #if we are on the menu, we update it
 
-    def removeMenu(self, name):
-        if name in self.name_list:
-            self.name_list.remove(name)
-        if name in self.menus.keys():
-            del self.menus[name]
-        self.selected = 0
-        self._showSelected()
+    def removeMenu(self, menu_id):
+        del self.menu_items[menu_id]
+        if self.selected == menu_id:
+            try:
+                self.selected = self.menu_items.iterkeys().next()
+            except StopIteration:
+                self.selected = None
+            self._showSelected()
 
     def checkShortcuts(self, key):
-        for menu in self.name_list:
-            key = self.menus[menu].checkShortcuts(key)
+        for menu_item in self.menu_items.values():
+            key = menu_item.widget.checkShortcuts(key)
         return key
 
 
@@ -949,29 +1129,35 @@ class GenericDialog(urwid.WidgetWrap):
             style = []
         frame_header = urwid.AttrMap(urwid.Text(title,'center'),'title')
 
-        buttons = None
+        self.buttons = collections.OrderedDict()
 
         if "OK/CANCEL" in style:
-            cancel_arg = [kwargs['cancel_value']] if kwargs.has_key('cancel_value') else []
-            ok_arg = [kwargs['ok_value']] if kwargs.has_key('ok_value') else []
-            buttons = [urwid.Button(_("Cancel"), kwargs['cancel_cb'], *cancel_arg),
-                      urwid.Button(_("Ok"), kwargs['ok_cb'], *ok_arg)]
+            self.buttons['cancel'] = urwid.Button(_("Cancel"), kwargs.get('cancel_cb'), kwargs.get('cancel_value'))
+            self.buttons['ok'] = urwid.Button(_("Ok"), kwargs.get('ok_cb'), kwargs.get('ok_value'))
         elif "YES/NO" in style:
-            yes_arg = [kwargs['yes_value']] if kwargs.has_key('yes_value') else []
-            no_arg = [kwargs['no_value']] if kwargs.has_key('no_value') else []
-            buttons = [urwid.Button(_("Yes"), kwargs['yes_cb'], *yes_arg),
-                      urwid.Button(_("No"), kwargs['no_cb'], *no_arg)]
+            self.buttons['yes'] = urwid.Button(_("Yes"), kwargs.get('yes_cb'), kwargs.get('yes_value'))
+            self.buttons['no'] = urwid.Button(_("No"), kwargs.get('no_cb'), kwargs.get('no_value'))
         if "OK" in style:
-            ok_arg = [kwargs['ok_value']] if kwargs.has_key('ok_value') else []
-            buttons = [urwid.Button(_("Ok"), kwargs['ok_cb'], *ok_arg)]
-        if buttons:
-            buttons_flow = urwid.GridFlow(buttons, max([len(button.get_label()) for button in buttons])+4, 1, 1, 'center')
+            self.buttons['ok'] = urwid.Button(_("Ok"), kwargs.get('ok_cb'), kwargs.get('ok_value'))
+        if self.buttons:
+            buttons_flow = urwid.GridFlow(self.buttons.values(), max([len(button.get_label()) for button in self.buttons.itervalues()])+4, 1, 1, 'center')
         body_content = urwid.SimpleListWalker(widgets_lst)
         frame_body = UnselectableListBox(body_content)
-        frame = FocusFrame(frame_body, frame_header, buttons_flow if buttons else None, 'footer' if buttons else 'body')
+        frame = FocusFrame(frame_body, frame_header, buttons_flow if self.buttons else None, 'footer' if self.buttons else 'body')
         decorated_frame = urwid.LineBox(frame)
         urwid.WidgetWrap.__init__(self, decorated_frame)
 
+    def setCallback(self, name, callback, data=None):
+        """Set the callback associated with a button press
+
+        @param name: one of "ok", "cancel", "yest", "no"
+        @aram callback(callable): method to call on requested action
+        @param data: argument to send to the callback (first one will be the button widget)
+        @raise KeyError if name is invalid
+        """
+        button = self.buttons[name]
+        urwid.connect_signal(button, 'click', callback, data)
+
 
 class InputDialog(GenericDialog):
     """Dialog with an edit box"""
@@ -1001,7 +1187,9 @@ class ConfirmDialog(GenericDialog):
 class Alert(GenericDialog):
     """Dialog with just a message and a OK button"""
 
-    def __init__(self, title, message, style=['OK'], **kwargs):
+    def __init__(self, title, message, style=None, **kwargs):
+        if style is None:
+            style= ['OK']
         GenericDialog.__init__(self, [urwid.Text(message, 'center')], title, style, ok_value=None, **kwargs)
 
 ## CONTAINERS ##
@@ -1260,23 +1448,28 @@ class TabsContainer(urwid.WidgetWrap):
         if not invisible:
             self._emit('click')
 
-    def _appendButton(self, name):
-        """Append a button to the frame header,
-        and link it to the page change method"""
+    def _appendButton(self, name, selected=False):
+        """Append a button to the frame header, and link it to the page change method.
+
+        @param name (unicode): button name
+        @param selected (bool): set to True to select this tab
+        """
         button = CustomButton(name, self._buttonClicked, left_border = '', right_border=' | ')
         self._buttons_cont.addWidget(button, button.getSize())
-        if len(self._buttons_cont.widget_list) == 1:
-            #first button: we set the focus and the body
-            self._buttons_cont.focus_position = 0
-            self._buttonClicked(button,True)
+        count = len(self._buttons_cont.widget_list)
+        if selected or count == 1:
+            # first/selected button: we set the focus and the body
+            self.selectTab(count - 1)
 
-    def addTab(self,name,content=None):
+    def addTab(self, name, content=None, selected=False):
         """Add a page to the container
+
         @param name: name of the page (what appear on the tab)
         @param content: content of the page:
             - if None create and empty Listbox
             - if it is a list instance, create a ListBox with the list in a body
             - else it must be a box widget which will be used instead of the ListBox
+        @param selected (bool): set to True to select this tab
         @return: ListBox (content of the page)"""
         if content is None or isinstance(content, list):
             tab = urwid.ListBox(urwid.SimpleListWalker(content or []))
@@ -1284,7 +1477,7 @@ class TabsContainer(urwid.WidgetWrap):
             tab = content
 
         self.tabs.append([name, tab])
-        self._appendButton(name)
+        self._appendButton(name, selected)
         return tab
 
     def addFooter(self, widget):
@@ -1292,6 +1485,14 @@ class TabsContainer(urwid.WidgetWrap):
         @param widget: FlowWidget"""
         self._w.footer = widget
 
+    def selectTab(self, index):
+        """Select a tab.
+
+        @param index (int): index of the tab to select
+        """
+        self._buttons_cont.focus_position = index
+        self._buttonClicked(self._buttons_cont.widget_list[index][1], True)
+
 
 class HighlightColumns(urwid.AttrMap):
     """ Decorated columns which highlight all or some columns """
